"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const parser_1 = require("../../parser");
const css_nodes_1 = require("../../utils/css-nodes");
function getInvalidEOFError(context, style) {
    var _a, _b;
    const node = context.getSourceCode().ast;
    const body = node.templateBody;
    let errors = (_a = body) === null || _a === void 0 ? void 0 : _a.errors;
    let inDocumentFragment = false;
    if (errors == null) {
        if (!context.parserServices.getDocumentFragment) {
            return null;
        }
        const df = context.parserServices.getDocumentFragment();
        inDocumentFragment = true;
        errors = (_b = df) === null || _b === void 0 ? void 0 : _b.errors;
        if (errors == null) {
            return null;
        }
    }
    const error = errors.find(err => typeof err.code === "string" &&
        err.code.startsWith("eof-") &&
        style.range[0] <= err.index &&
        err.index < style.range[1]) ||
        errors.find(err => typeof err.code === "string" && err.code.startsWith("eof-"));
    if (!error) {
        return null;
    }
    return {
        error,
        inDocumentFragment,
    };
}
function getStyleElements(context) {
    let document = null;
    if (context.parserServices.getDocumentFragment) {
        document = context.parserServices.getDocumentFragment();
    }
    else {
        const sourceCode = context.getSourceCode();
        const { ast } = sourceCode;
        const templateBody = ast.templateBody;
        if (templateBody) {
            document = templateBody.parent;
        }
    }
    if (document) {
        return document.children
            .filter(isVElement)
            .filter(element => element.name === "style");
    }
    return [];
}
function isScoped(style) {
    const { startTag } = style;
    return startTag.attributes.some(attr => attr.key.name === "scoped");
}
function getLang(style) {
    var _a, _b;
    const { startTag } = style;
    const lang = startTag.attributes.find(attr => attr.key.name === "lang") || null;
    return (((_a = lang) === null || _a === void 0 ? void 0 : _a.type) === "VAttribute" &&
        ((_b = lang.value) === null || _b === void 0 ? void 0 : _b.type) === "VLiteral" &&
        lang.value.value);
}
class StyleContext {
    constructor(style, context) {
        const sourceCode = context.getSourceCode();
        this.styleElement = style;
        this.sourceCode = sourceCode;
        const { startTag, endTag } = style;
        this.invalid = null;
        const eof = getInvalidEOFError(context, style);
        if (eof) {
            this.invalid = {
                message: eof.error.message,
                needReport: eof.inDocumentFragment,
                loc: { line: eof.error.lineNumber, column: eof.error.column },
            };
        }
        else if (endTag == null) {
            this.invalid = {
                message: "Missing end tag",
                needReport: true,
                loc: startTag.loc.end,
            };
        }
        this.scoped = Boolean(style && isScoped(style));
        this.lang = ((style && getLang(style)) || "css").toLowerCase();
        if (!this.invalid && endTag != null) {
            this.cssText = sourceCode.text.slice(startTag.range[1], endTag.range[0]);
            this.cssNode = parser_1.parse(sourceCode, startTag.loc.end, this.cssText, this.lang);
        }
        else {
            this.cssText = null;
            this.cssNode = null;
        }
    }
    traverseNodes(visitor) {
        if (this.cssNode) {
            traverseNodes(this.cssNode, visitor);
        }
    }
}
exports.StyleContext = StyleContext;
function traverseNodes(node, visitor) {
    visitor.break = false;
    visitor.enterNode(node);
    if (visitor.exit || visitor.break) {
        return;
    }
    if (css_nodes_1.isVCSSContainerNode(node)) {
        for (const child of node.nodes) {
            traverseNodes(child, visitor);
            if (visitor.break) {
                break;
            }
            if (visitor.exit) {
                return;
            }
        }
    }
    visitor.leaveNode(node);
}
function createStyleContexts(context) {
    const styles = getStyleElements(context);
    return styles.map(style => new StyleContext(style, context));
}
exports.createStyleContexts = createStyleContexts;
function isVElement(node) {
    var _a;
    return ((_a = node) === null || _a === void 0 ? void 0 : _a.type) === "VElement";
}
