"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const css_selector_parser_1 = require("./css-selector-parser");
const postcss_selector_parser_1 = __importDefault(require("postcss-selector-parser"));
const lodash_1 = __importDefault(require("lodash"));
const utils_1 = require("../utils");
const ast_1 = require("../../ast");
class RemapContext {
    constructor() {
        this.mappers = [];
        this.orgIndex = 0;
        this.newIndex = 0;
        this.batchLengthOrg = 0;
        this.batchLengthNew = 0;
    }
    applyEq(length) {
        if (length <= 0) {
            return;
        }
        this.flush();
        const newEnd = this.newIndex + length;
        const orgEnd = this.orgIndex + length;
        this.addMap([this.orgIndex, orgEnd], [this.newIndex, newEnd]);
        this.newIndex = newEnd;
        this.orgIndex = orgEnd;
    }
    applyIns(length) {
        this.batchLengthNew += length;
    }
    applyDel(length) {
        this.batchLengthOrg += length;
    }
    flush() {
        if (this.batchLengthNew || this.batchLengthOrg) {
            const newEnd = this.newIndex + this.batchLengthNew;
            const orgEnd = this.orgIndex + this.batchLengthOrg;
            this.addMap([this.orgIndex, orgEnd], [this.newIndex, newEnd]);
            this.newIndex = newEnd;
            this.orgIndex = orgEnd;
            this.batchLengthOrg = 0;
            this.batchLengthNew = 0;
        }
    }
    addMap(orgRange, newRange) {
        if (orgRange[0] === newRange[0] && orgRange[1] === newRange[1]) {
            return;
        }
        this.mappers.unshift({
            org: orgRange,
            new: newRange,
        });
    }
    hasMapping() {
        return this.mappers.length > 1;
    }
    remapIndex(index) {
        for (const mapper of this.mappers) {
            if (mapper.new[0] <= index && index < mapper.new[1]) {
                const offset = index - mapper.new[0];
                return Math.min(mapper.org[0] + offset, mapper.org[1] - 1);
            }
            if (index === mapper.new[1]) {
                return mapper.org[1];
            }
        }
        return index;
    }
}
class SourceCode {
    constructor(code) {
        const lineStartIndices = [0];
        let match = undefined;
        const lineEndingPattern = /\r\n|[\r\n\u2028\u2029]/gu;
        while ((match = lineEndingPattern.exec(code))) {
            lineStartIndices.push(match.index + match[0].length);
        }
        this.text = code;
        this.lineStartIndices = lineStartIndices;
    }
    getLocFromIndex(index) {
        const code = this.text;
        const lineStartIndices = this.lineStartIndices;
        if (index === code.length) {
            return {
                line: lineStartIndices.length,
                column: index - lineStartIndices[lineStartIndices.length - 1],
            };
        }
        const lineNumber = lodash_1.default.sortedLastIndex(lineStartIndices, index);
        return {
            line: lineNumber,
            column: index - lineStartIndices[lineNumber - 1],
        };
    }
    getIndexFromLoc(loc) {
        const lineStartIndices = this.lineStartIndices;
        const lineStartIndex = lineStartIndices[loc.line - 1];
        const positionIndex = lineStartIndex + loc.column;
        return positionIndex;
    }
}
class SCSSSelectorParser extends css_selector_parser_1.CSSSelectorParser {
    parseInternal(selector) {
        const { cssSelector, remapContext, interpolations, inlineComments, } = replaceSelectorInterpolationAndInlineComments(selector);
        const result = postcss_selector_parser_1.default().astSync(cssSelector);
        if (!interpolations.length && !inlineComments.length) {
            return result;
        }
        const scssSourceCode = new SourceCode(selector);
        const cssSourceCode = new SourceCode(cssSelector);
        return restoreInterpolationAndInlineComments(result, {
            remapContext,
            interpolations,
            inlineComments,
            scssSourceCode,
            cssSourceCode,
        });
    }
    parseCommentsInternal(selector) {
        return this.parseInternal(selector);
    }
    convertCommentNode(node, loc, start, end, parent) {
        if (node.value.startsWith("//")) {
            const text = node.value
                .replace(/^\s*\/\//u, "")
                .replace(/\r?\n\s*$/u, "");
            this.commentContainer.push(new ast_1.VCSSInlineComment(node, text, loc, start, end, {
                parent,
            }));
            return null;
        }
        return super.convertCommentNode(node, loc, start, end, parent);
    }
}
exports.SCSSSelectorParser = SCSSSelectorParser;
function replaceSelectorInterpolationAndInlineComments(selector) {
    const remapContext = new RemapContext();
    const interpolations = new Map();
    const inlineComments = new Map();
    const reInterpolationOrComment = /#\{([\s\S]+?)\}|\/\*([\s\S]+?)\*\/|\/\/([\s\S]+?)\n/gu;
    let res = null;
    let cssSelector = "";
    let start = 0;
    while ((res = reInterpolationOrComment.exec(`${selector}\n`))) {
        const text = res[0];
        if (text.startsWith("/*")) {
            continue;
        }
        const plain = selector.slice(start, res.index);
        let replaceString;
        if (text.startsWith("//")) {
            let block = `/*${randomStr()}*/`;
            while (inlineComments.has(block) || selector.includes(block)) {
                block = `/*${randomStr()}*/`;
            }
            inlineComments.set(block, {
                start: res.index,
                original: text,
                replace: block,
            });
            replaceString = block;
        }
        else {
            let interpolate = `_${randomStr()}_`;
            while (interpolations.has(interpolate) ||
                selector.includes(interpolate)) {
                interpolate = `_${randomStr()}_`;
            }
            interpolations.set(interpolate, {
                start: res.index,
                original: text,
                replace: interpolate,
            });
            replaceString = interpolate;
        }
        cssSelector += plain;
        cssSelector += replaceString;
        remapContext.applyEq(plain.length);
        remapContext.applyIns(replaceString.length);
        remapContext.applyDel(res[0].length);
        start = reInterpolationOrComment.lastIndex;
    }
    const plain = selector.slice(start);
    cssSelector += plain;
    remapContext.applyEq(plain.length);
    remapContext.flush();
    return {
        cssSelector,
        remapContext,
        interpolations: [...interpolations.values()],
        inlineComments: [...inlineComments.values()],
    };
}
function restoreInterpolationAndInlineComments(node, options) {
    const { remapContext, interpolations, inlineComments, scssSourceCode, cssSourceCode, } = options;
    if (node.source) {
        if (node.source.start) {
            const cssLoc = node.source.start;
            const index = cssSourceCode.getIndexFromLoc({
                line: cssLoc.line,
                column: cssLoc.column - 1,
            });
            const scssIndex = remapContext.remapIndex(index);
            const scssLoc = scssSourceCode.getLocFromIndex(scssIndex);
            scssLoc.column++;
            node.source.start = scssLoc;
        }
        if (node.source.end) {
            const cssLoc = node.source.end;
            const index = cssSourceCode.getIndexFromLoc({
                line: cssLoc.line,
                column: cssLoc.column,
            });
            const scssIndex = remapContext.remapIndex(index);
            node.source.end = scssSourceCode.getLocFromIndex(scssIndex);
        }
        while (restoreInterpolations(node, interpolations)) {
        }
        while (restoreComments(node, inlineComments)) {
        }
    }
    if (utils_1.isPostCSSSPContainer(node)) {
        for (const n of node.nodes) {
            restoreInterpolationAndInlineComments(n, options);
        }
    }
    return node;
}
function restoreInterpolations(node, interpolations) {
    if (!interpolations.length) {
        return false;
    }
    const targetProperties = [];
    if (node.type === "tag" ||
        node.type === "class" ||
        node.type === "id" ||
        node.type === "combinator" ||
        node.type === "pseudo" ||
        node.type === "string") {
        targetProperties.push("value");
    }
    else if (node.type === "attribute") {
        targetProperties.push("attribute");
        targetProperties.push("value");
    }
    for (const prop of targetProperties) {
        for (let index = 0; index < interpolations.length; index++) {
            const interpolation = interpolations[index];
            if (restoreReplaceNodeProp(node, prop, interpolation)) {
                interpolations.splice(index, 1);
                return true;
            }
        }
    }
    return false;
}
function restoreComments(node, inlineComments) {
    if (!inlineComments.length) {
        return false;
    }
    const targetProperties = [];
    if (node.type === "comment") {
        targetProperties.push("value");
    }
    if (hasRaws(node)) {
        targetProperties.push("raws.spaces.after");
        targetProperties.push("raws.spaces.before");
    }
    for (const prop of targetProperties) {
        for (let index = 0; index < inlineComments.length; index++) {
            const comment = inlineComments[index];
            if (restoreReplaceNodeProp(node, prop, comment)) {
                inlineComments.splice(index, 1);
                return true;
            }
        }
    }
    return false;
}
function restoreReplaceNodeProp(node, prop, replaceInfo) {
    const text = `${lodash_1.default.get(node, prop, "") || ""}`;
    if (text.includes(replaceInfo.replace)) {
        const newText = text.replace(replaceInfo.replace, replaceInfo.original);
        lodash_1.default.set(node, prop, newText);
        if (!prop.startsWith("raws")) {
            lodash_1.default.set(node, `raws.${prop}`, newText);
        }
        return true;
    }
    return false;
}
function randomStr() {
    const S = "abcdefghijklmnopqrstuvwxyz0123456789";
    const N = 16;
    return Array.from(Array(N))
        .map(() => S[Math.floor(Math.random() * S.length)])
        .join("");
}
function hasRaws(node) {
    return node.raws != null;
}
