"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const selectors_1 = require("../../utils/selectors");
const css_nodes_1 = require("../../utils/css-nodes");
class ResolvedSelector {
    constructor(selector, container) {
        this.selector = selector;
        this.container = container;
    }
}
exports.ResolvedSelector = ResolvedSelector;
class CSSSelectorResolver {
    resolveSelectors(rootStyle) {
        return this.resolveNodesSelectors(rootStyle.nodes, null);
    }
    resolveNodesSelectors(nodes, parentSelectors) {
        const results = [];
        for (const node of nodes) {
            const selectorNodes = this.getSelectorNodes(node);
            let selectors = null;
            if (selectorNodes) {
                selectors = [];
                if (!parentSelectors) {
                    for (const selectorNode of selectorNodes.filter(selectors_1.hasNodesSelector)) {
                        const rootResolvedSelector = this.createRootResolvedSelector(selectorNode.nodes, node);
                        results.push(rootResolvedSelector);
                        selectors.push(rootResolvedSelector);
                    }
                }
                else {
                    for (const selectorNode of selectorNodes.filter(selectors_1.hasNodesSelector)) {
                        const resolvedSelectors = this.resolveNestingSelectors(selectorNode.nodes, parentSelectors, node);
                        for (const resolvedSelector of resolvedSelectors) {
                            results.push(resolvedSelector);
                            selectors.push(resolvedSelector);
                        }
                    }
                }
            }
            else if (this.isIgnoreNode(node)) {
                continue;
            }
            else {
                selectors = parentSelectors;
            }
            if (css_nodes_1.isVCSSContainerNode(node)) {
                results.push(...this.resolveNodesSelectors(node.nodes, selectors));
            }
        }
        return results;
    }
    getSelectorNodes(node) {
        if (css_nodes_1.isVCSSStyleRule(node) || selectors_1.isNestingAtRule(node)) {
            return node.selectors;
        }
        return null;
    }
    isIgnoreNode(node) {
        return css_nodes_1.isVCSSAtRule(node) && node.name === "keyframes";
    }
    createRootResolvedSelector(selectorNodes, container) {
        return new ResolvedSelector(selectorNodes, container);
    }
    resolveNestingSelectors(selectorNodes, parentSelectors, container) {
        if (selectors_1.isNestingAtRule(container)) {
            return this.resolveSelectorForNestContaining(selectorNodes, parentSelectors, container);
        }
        const nestingIndex = selectorNodes.findIndex(selectors_1.isNestingSelector);
        if (nestingIndex === 0) {
            return this.resolveSelectorForNestPrefixed(selectorNodes, parentSelectors, container);
        }
        return [new ResolvedSelector(selectorNodes, container)];
    }
    resolveSelectorForNestPrefixed(selectorNodes, parentSelectors, container) {
        if (!selectorNodes.length || !selectors_1.isNestingSelector(selectorNodes[0])) {
            return [new ResolvedSelector(selectorNodes, container)];
        }
        const after = selectorNodes.slice(1);
        return parentSelectors.map(parentSelector => new ResolvedSelector([...parentSelector.selector, ...after], container));
    }
    resolveSelectorForNestContaining(selectorNodes, parentSelectors, container) {
        const nesting = selectors_1.findNestingSelectors(selectorNodes).next().value;
        if (!nesting) {
            return [new ResolvedSelector(selectorNodes, container)];
        }
        const { nestingIndex, nodes: hasNestingNodes, node: nestingNode, } = nesting;
        const nestingLeftNode = hasNestingNodes[nestingIndex - 1];
        const nestingRightNode = hasNestingNodes[nestingIndex + 1];
        const maybeJoinLeft = nestingLeftNode && nestingLeftNode.range[1] === nestingNode.range[0];
        const needJoinRight = nestingRightNode &&
            nestingNode.range[1] === nestingRightNode.range[0] &&
            selectors_1.isTypeSelector(nestingRightNode);
        const beforeSelector = hasNestingNodes.slice(0, nestingIndex);
        const afterSelector = hasNestingNodes.slice(nestingIndex + 1);
        let resolved = parentSelectors.map(p => {
            const before = [...beforeSelector];
            const after = [...afterSelector];
            const parentSelector = [...p.selector];
            const needJoinLeft = maybeJoinLeft && selectors_1.isTypeSelector(parentSelector[0]);
            if (needJoinLeft && needJoinRight && parentSelector.length === 1) {
                before.push(newNestingConcatBothSelectorNodes(before.pop(), parentSelector.shift(), after.shift(), nestingNode));
            }
            else {
                if (needJoinLeft) {
                    before.push(newNestingConcatLeftSelectorNodes(before.pop(), parentSelector.shift(), nestingNode));
                }
                if (needJoinRight) {
                    after.unshift(newNestingConcatRightSelectorNodes(parentSelector.pop(), after.shift(), nestingNode));
                }
            }
            return new ResolvedSelector([...before, ...parentSelector, ...after], container);
        });
        let nestingTargetNode = nestingNode;
        while (!selectorNodes.includes(nestingTargetNode)) {
            const parent = nestingTargetNode.parent;
            const index = parent.parent.nodes.indexOf(parent);
            const before = parent.parent.nodes.slice(0, index);
            const after = parent.parent.nodes.slice(index + 1);
            resolved = resolved.map(selector => {
                const newNode = parent.copy();
                newNode.nodes = selector.selector;
                return new ResolvedSelector([...before, newNode, ...after], container);
            });
            nestingTargetNode = parent;
        }
        return resolved;
    }
}
exports.CSSSelectorResolver = CSSSelectorResolver;
function newNestingConcatBothSelectorNodes(left, parent, right, _nesting) {
    const loc = {
        start: left.loc.start,
        end: right.loc.end,
    };
    const newNode = left.copy({
        loc,
        start: left.range[0],
        end: right.range[1],
        parent: left.parent,
        value: `${left.value}${parent.selector}${right.selector}`,
    });
    return newNode;
}
function newNestingConcatLeftSelectorNodes(left, parent, nesting) {
    const loc = {
        start: left.loc.start,
        end: nesting.loc.end,
    };
    const newNode = left.copy({
        loc,
        start: left.range[0],
        end: nesting.range[1],
        parent: left.parent,
        value: `${left.value}${parent.selector}`,
    });
    return newNode;
}
function newNestingConcatRightSelectorNodes(parent, right, nesting) {
    const loc = {
        start: nesting.loc.start,
        end: right.loc.end,
    };
    const newNode = parent.copy({
        node: right.node,
        loc,
        start: nesting.range[0],
        end: right.range[1],
        parent: right.parent,
        value: `${parent.value}${right.selector}`,
    });
    return newNode;
}
